import typing as t

import requests
from praw.models import Submission
from rich.status import Status

from .client import reddit


class Posts:
    """Represents Reddit posts and provides methods for retrieving posts from various sources."""

    @classmethod
    def best(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Submission]:
        posts = reddit.posts(
            session=session,
            status=status,
            kind="best",
            limit=limit,
            sort="all",
        )

        return posts

    @classmethod
    def controversial(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Submission]:
        posts = reddit.posts(
            session=session,
            status=status,
            kind="controversial",
            limit=limit,
            sort="all",
        )

        return posts

    @classmethod
    def front_page(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Submission]:
        posts = reddit.posts(
            session=session,
            status=status,
            kind="front_page",
            limit=limit,
        )

        return posts

    @classmethod
    def new(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Submission]:

        posts = reddit.posts(
            session=session,
            status=status,
            kind="new",
            limit=limit,
        )

        return posts

    @classmethod
    def top(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Submission]:
        posts = reddit.posts(
            session=session,
            status=status,
            kind="top",
            limit=limit,
            sort="all",
        )

        return posts

    @classmethod
    def rising(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Submission]:
        posts = reddit.posts(
            session=session,
            status=status,
            kind="rising",
            limit=limit,
            sort="all",
        )

        return posts



import typing as t

import requests
from praw.models import Submission, Subreddit, Redditor
from rich.status import Status

from .client import reddit


class Search:
    def __init__(self, query: str):
        self._query = query

    def posts(
        self,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Submission]:
        search_results = reddit.search(
            session=session,
            status=status,
            kind="posts",
            query=self._query,
            limit=limit,
        )

        return search_results

    def subreddits(
        self,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Subreddit]:
        search_results = reddit.search(
            session=session,
            status=status,
            kind="subreddits",
            query=self._query,
            limit=limit,
        )

        return search_results

    def users(
        self,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Redditor]:
        search_results = reddit.search(
            session=session,
            status=status,
            kind="users",
            query=self._query,
            limit=limit,
        )

        return search_results



import typing as t
from collections import Counter

from praw.models import Submission, Redditor, Comment, Subreddit
from rich.status import Status

from karmakrate.riches import rich_colours
from karmakrate.riches.rich_logging import console
from karmakrate.riches.rich_render import Render
from .client import reddit, LISTINGS


class User:
    def __init__(self, username: str):
        self._redditor = reddit.redditor(name=username)
        self._username = username

    def comments(
        self,
        limit: int,
        listing: LISTINGS,
        status: t.Optional[Status] = None,
    ) -> t.Union[t.List[Comment], None]:
        if self.exists(status=status):
            if isinstance(status, Status):
                status.update(
                    f"Getting {limit} {listing} comments from u/{self._username}..."
                )

            func = getattr(self._redditor, listing)
            return list(func(limit=limit))

        else:
            return None

    def moderated(
        self, status: t.Optional[Status] = None
    ) -> t.Union[t.List[Subreddit], None]:
        if self.exists(status=status):
            if isinstance(status, Status):
                self.exists(status=status)

                status.update(
                    f"Getting moderated subreddits from u/{self._username}..."
                )
            return list(self._redditor.moderated())

        else:
            return None

    def overview(
        self,
        status: t.Optional[Status] = None,
    ) -> t.Union[t.List[Comment], None]:
        if self.exists(status=status):
            if isinstance(status, Status):
                status.update(f"Getting recent comments from u/{self._username}...")

            return list(self._redditor.comments.new(limit=None))

        else:
            return None

    def posts(
        self,
        limit: t.Optional[int],
        listing: LISTINGS,
        status: t.Optional[Status] = None,
    ) -> t.Union[t.List[Submission], None]:
        if self.exists(status=status):
            if isinstance(status, Status):
                status.update(
                    f"Getting {limit} {listing} posts from u/{self._username}..."
                )

            func = getattr(self._redditor.submissions, listing)
            return list(func(limit=limit))
        else:
            return None

    def profile(
        self,
        status: t.Optional[Status] = None,
    ) -> t.Union[Redditor, None]:
        if self.exists(status=status):
            if isinstance(status, Status):
                status.update(f"Getting profile info from u/{self._username}...")

            return self._redditor

        else:
            return None

    def top_subreddits(
        self,
        top_n: int,
        status: t.Optional[Status] = None,
    ):
        if self.exists(status=status):
            posts = self.posts(status=status, limit=None, listing="top")

            if posts:
                # Extract subreddit names
                subreddits = [post.data.subreddit for post in posts]

                # Count the occurrences of each subreddit
                subreddit_counts: t.Counter = Counter(subreddits)

                # Get the most common subreddits
                top_subreddits: t.List[tuple[str, int]] = subreddit_counts.most_common(
                    top_n
                )

                # Prepare data for plotting
                subreddit_names = [subreddit[0] for subreddit in top_subreddits]
                subreddit_frequencies = [subreddit[1] for subreddit in top_subreddits]

                data: t.Dict[str, int] = dict(
                    zip(subreddit_names, subreddit_frequencies)
                )

                Render.bar_chart(
                    data=data,
                    title=f"top {top_n}/{len(posts)} subreddits analysis",
                    x_label="Subreddits",
                    y_label="Frequency",
                )

    def exists(
        self,
        status: t.Optional[Status] = None,
    ) -> bool:
        if isinstance(status, Status):
            status.update(f"Checking user availability...")

        verdict: bool = (
            True if not reddit.username_available(name=self._username) else False
        )

        if verdict:
            console.print(
                f"{rich_colours.BOLD_GREEN}✔{rich_colours.BOLD_GREEN_RESET} {self._username} is a real user"
            )

        elif not verdict:
            console.print(
                f"{rich_colours.BOLD_YELLOW}✘{rich_colours.BOLD_YELLOW_RESET} {self._username} is not a real user"
            )
        return verdict



import typing as t

from praw.models import Submission, Comment
from praw.models.reddit.subreddit import SubredditWiki
from prawcore import exceptions
from rich.status import Status

from karmakrate.riches import rich_colours
from karmakrate.riches.rich_logging import console
from .client import reddit, TIME_FILTERS, SORT, LISTINGS
from .shared import is_empty_data


class Subreddit:
    def __init__(self, display_name: str):
        self._display_name = display_name
        self._subreddit = reddit.subreddit(display_name=display_name)

    def comments(
        self, limit: int, status: t.Optional[Status] = None
    ) -> t.Union[t.List[Comment], None]:
        """
        Retrieves a list of comments from the subreddit.

        :param limit: Maximum number of comments to retrieve.
        :type limit: int
        :param status: Optional status object for updating progress.
        :type status: t.Optional[Status]
        :return: List of comments from the subreddit, or None if the subreddit does not exist.
        :rtype: t.Union[t.List[Comment], None]
        """
        if self.exists(status=status):
            if isinstance(status, Status):
                status.update(
                    f"Getting {limit} comments from {self._subreddit.display_name_prefixed}..."
                )
            comments = [
                comment.refresh() for comment in self._subreddit.comments(limit=limit)
            ]
            return is_empty_data(
                data=comments,
                message=f"No comments found in {self._subreddit.display_name_prefixed}.",
            )
        else:
            return None

    def posts(
        self, limit: int, listing: LISTINGS, status: t.Optional[Status] = None
    ) -> t.Union[t.List[Submission], None]:
        """
        Retrieves a list of posts from the subreddit based on the specified listing type.

        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param listing: Type of listing to retrieve (e.g., 'hot', 'new', 'top').
        :param status: Optional status object for updating progress.
        :type status: t.Optional[Status]
        :return: List of posts from the subreddit, or None if the subreddit does not exist.
        :rtype: t.Union[t.List[Submission], None]
        """
        if self.exists(status=status):
            if isinstance(status, Status):
                status.update(
                    f"Getting {limit} {listing} posts from {self._subreddit.display_name_prefixed}..."
                )
            func = getattr(self._subreddit, listing)
            return is_empty_data(
                data=list(func(limit=limit)),
                message=f"No {listing} posts found in {self._subreddit.display_name_prefixed}.",
            )
        else:
            return None

    def profile(self, status: t.Optional[Status] = None) -> t.Union["Subreddit", None]:
        if self.exists(status=status):
            if isinstance(status, Status):
                status.update(
                    f"Getting profile data from subreddit r/{self._display_name}..."
                )

            return self._subreddit
        else:
            return None

    def search(
        self,
        query: str,
        limit: int,
        sort: SORT,
        time_filter: TIME_FILTERS,
        status: t.Optional[Status] = None,
    ) -> t.Union[t.List[Submission], None]:
        """
        Searches for posts in the subreddit based on the provided query.

        :param query: Search query string.
        :type query: str
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sorting method for the search results.
        :type sort: SORT
        :param time_filter: Time filter for the search results.
        :type time_filter: TIME_FILTERS
        :param status: Optional status object for updating progress.
        :type status: t.Optional[Status]
        :return: List of posts matching the search query, or None if the subreddit does not exist.
        :rtype: t.Union[t.List[Submission], None]
        """
        if self.exists(status=status):
            if isinstance(status, Status):
                status.update(
                    f"Searching for '{query}' in posts from {self._subreddit.display_name_prefixed}..."
                )
            results = self._subreddit.search(
                query=query,
                limit=limit,
                sort=sort,
                time_filter=time_filter,
            )

            return is_empty_data(
                data=list(results),
                message=f"No results found for '{query}' in {self._subreddit.display_name_prefixed}.",
            )
        else:
            return None

    def wiki_pages(
        self, status: t.Optional[Status] = None
    ) -> t.Union[t.List[SubredditWiki], None]:
        """
        Retrieves a list of wiki pages from the subreddit.

        :param status: Optional status object for updating progress.
        :type status: t.Optional[Status]
        :return: List of wiki pages from the subreddit, or None if the subreddit does not exist.
        :rtype: t.Union[t.List[SubredditWiki], None]
        """
        if self.exists(status=status):
            if isinstance(status, Status):
                status.update(
                    f"Getting wiki pages from {self._subreddit.display_name_prefixed}...",
                )

            pages = self._subreddit.wiki
            return is_empty_data(
                data=list(pages),
                message=f"No wiki pages found in {self._subreddit.display_name_prefixed}.",
            )
        else:
            return None

    def exists(self, status: t.Optional[Status] = None) -> bool:
        """
        Checks if the subreddit exists by attempting to access its ID.

        :param status: Optional status object for updating progress.
        :type status: t.Optional[Status]
        :return: True if the subreddit exists, False otherwise.
        :rtype: bool
        """
        if isinstance(status, Status):
            status.update(f"Checking subreddit availability...")

        try:
            _ = self._subreddit.id
            verdict = True
        except exceptions.Redirect:
            verdict = False
        except exceptions.NotFound:
            verdict = False
        except exceptions.Forbidden:
            verdict = False

        if verdict:
            console.print(
                f"{rich_colours.BOLD_GREEN}✔{rich_colours.BOLD_GREEN_RESET} {self._display_name} is a real subreddit"
            )

        elif not verdict:
            console.print(
                f"{rich_colours.BOLD_YELLOW}✘{rich_colours.BOLD_YELLOW_RESET} {self._display_name} is not a real subreddit"
            )
        return verdict



__all__ = ["Author", "Project"]


class Author:
    name: str = "Ritchie Mwewa"
    username: str = "rly0nheart"
    gravatar: str = f"https://gravatar.com/{username}"


class Project:
    name: str = "Knew Karma"
    package: str = "knewkarma"
    summary: str = f"Reddit-data analysis toolkit — by {Author.name}"
    description: str = f"""
{name} (/nuː ‘kɑːrmə/) is an analysis toolkit designed to provide an extensive range of
functionalities for exploring and analysing Reddit data. It includes a Command-Line Interface (CLI), and an
Application Programming Interface (API) to enable an easy integration in other Python Projects.
"""


# -------------------------------- END ----------------------------------------- #



import typing as t

from praw.models import Submission, Comment
from rich.status import Status

from .client import reddit


class Post:
    def __init__(self, id: str):
        self.id = id
        self._post = reddit.submission(id=id)

    def info(self, status: Status) -> Submission:
        if isinstance(status, Status):
            status.update(f"Getting info from post {self.id}...")
        return self._post

    def comments(self, status: Status) -> t.List[Comment]:
        if isinstance(status, Status):
            status.update(f"Getting comments from post {self.id}...")
        return self._post.comments.list()



import typing as t

from karmakrate.riches import rich_colours
from karmakrate.riches.rich_logging import console


def is_empty_data(data: t.List, message: str) -> t.List:
    """
    Check if given data is empty.

    :param data: List of data to check.
    :type data: t.List
    :param message: Message to print if data is empty.
    :type message: str
    :return: An empty list if data is empty, otherwise the original data.
    :rtype: t.List
    """

    if not data:
        console.print(
            f"{rich_colours.BOLD_YELLOW}✘{rich_colours.BOLD_YELLOW_RESET} {message}"
        )
        return []
    else:
        return data



import typing as t

import requests
from praw.models import Subreddit
from rich.status import Status

from .client import reddit


class Subreddits:
    @classmethod
    def all(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Subreddit]:
        all_subreddits = reddit.subreddits(
            session=session,
            status=status,
            kind="all",
            limit=limit,
        )

        return all_subreddits

    @classmethod
    def default(
        cls,
        limit: int,
        session: requests.Session,
        status: t.Optional[Status] = None,
    ) -> t.List[Subreddit]:
        default_subreddits = reddit.subreddits(
            session=session,
            status=status,
            kind="default",
            timeframe="all",
            limit=limit,
        )

        return default_subreddits

    @classmethod
    def new(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Subreddit]:
        new_subreddits = reddit.subreddits(
            session=session,
            status=status,
            kind="new",
            limit=limit,
        )

        return new_subreddits

    @classmethod
    def popular(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Subreddit]:
        popular_subreddits = reddit.subreddits(
            session=session,
            status=status,
            kind="popular",
            limit=limit,
        )

        return popular_subreddits



import typing as t
from platform import platform, python_version

import praw

from karmakrate.handlers.auth_handler import AuthHandler
from ..meta.about import Project
from ..meta.version import Version

USER_AGENT: str = (
    f"{Project.name.replace(' ', '-')}/{Version.release} "
    f"(Python {python_version} on {platform}; +{Project.documentation})"
)
LISTINGS = t.Literal["controversial", "gilded", "hot", "new", "rising", "top"]
TIME_FILTERS = t.Literal["all", "hour", "day", "week", "month", "year"]
SORT = t.Literal["relevance", "hot", "top", "new", "lucene", "all"]

reddit = praw.Reddit(
    client_id=AuthHandler.read()["client_id"],
    client_secret=AuthHandler.read()["client_secret"],
    user_agent=USER_AGENT,
)



from datetime import date

from karmakrate.riches import rich_colours
from .about import Author, Project

__all__ = ["License"]

from .version import Version

_current_year: int = date.today().year

_author = Author
_project = Project


class License:
    conditions: str = """
[bold]TERMS AND CONDITIONS[/]

0. Definitions.

    "This License" refers to version 3 of the GNU General Public License.
    
    "Copyright" also means copyright-like laws that apply to other kinds of
    works, such as semiconductor masks.
    
    "The Program" refers to any copyrightable work licensed under this
    License.  Each licensee is addressed as "you".  "Licensees" and
    "recipients" may be individuals or organizations.
    
    To "modify" a work means to copy from or adapt all or part of the work
    in a fashion requiring copyright permission, other than the making of an
    exact copy.  The resulting work is called a "modified version" of the
    earlier work or a work "based on" the earlier work.
    
    A "covered work" means either the unmodified Program or a work based
    on the Program.
    
    To "propagate" a work means to do anything with it that, without
    permission, would make you directly or secondarily liable for
    infringement under applicable copyright law, except executing it on a
    computer or modifying a private copy.  Propagation includes copying,
    distribution (with or without modification), making available to the
    public, and in some countries other activities as well.
    
    To "convey" a work means any kind of propagation that enables other
    parties to make or receive copies.  Mere interaction with a user through
    a computer network, with no transfer of a copy, is not conveying.
    
    An interactive user interface displays "Appropriate Legal Notices"
    to the extent that it includes a convenient and prominently visible
    feature that (1) displays an appropriate copyright notice, and (2)
    tells the user that there is no warranty for the work (except to the
    extent that warranties are provided), that licensees may convey the
    work under this License, and how to view a copy of this License.  If
    the interface presents a list of user commands or options, such as a
    menu, a prominent item in the list meets this criterion.

1. Source Code.

    The "source code" for a work means the preferred form of the work
    for making modifications to it.  "Object code" means any non-source
    form of a work.
    
    A "Standard Interface" means an interface that either is an official
    standard defined by a recognized standards body, or, in the case of
    interfaces specified for a particular programming language, one that
    is widely used among developers working in that language.
    
    The "System Libraries" of an executable work include anything, other
    than the work as a whole, that (a) is included in the normal form of
    packaging a Major Component, but which is not part of that Major
    Component, and (b) serves only to enable use of the work with that
    Major Component, or to implement a Standard Interface for which an
    implementation is available to the public in source code form.  A
    "Major Component", in this context, means a major essential component
    (kernel, window system, and so on) of the specific operating system
    (if any) on which the executable work runs, or a compiler used to
    produce the work, or an object code interpreter used to run it.
    
    The "Corresponding Source" for a work in object code form means all
    the source code needed to generate, install, and (for an executable
    work) run the object code and to modify the work, including scripts to
    control those activities.  However, it does not include the work's
    System Libraries, or general-purpose _utils or generally available free
    programs which are used unmodified in performing those activities but
    which are not part of the work.  For example, Corresponding Source
    includes interface definition files associated with source files for
    the work, and the source code for shared libraries and dynamically
    linked subprograms that the work is specifically designed to require,
    such as by intimate data communication or control flow between those
    subprograms and other parts of the work.
    
    The Corresponding Source need not include anything that users
    can regenerate automatically from other parts of the Corresponding
    Source.
    
    The Corresponding Source for a work in source code form is that
    same work.

2. Basic Permissions.

    All rights granted under this License are granted for the term of
    copyright on the Program, and are irrevocable provided the stated
    conditions are met.  This License explicitly affirms your unlimited
    permission to run the unmodified Program.  The output from running a
    covered work is covered by this License only if the output, given its
    content, constitutes a covered work.  This License acknowledges your
    rights of fair use or other equivalent, as provided by copyright law.
    
    You may make, run and propagate covered works that you do not
    convey, without conditions so long as your license otherwise remains
    in force.  You may convey covered works to others for the sole purpose
    of having them make modifications exclusively for you, or provide you
    with facilities for running those works, provided that you comply with
    the terms of this License in conveying all material for which you do
    not control copyright.  Those thus making or running the covered works
    for you must do so exclusively on your behalf, under your direction
    and control, on terms that prohibit them from making any copies of
    your copyrighted material outside their relationship with you.
    
    Conveying under any other circumstances is permitted solely under
    the conditions stated below.  Sublicensing is not allowed; section 10
    makes it unnecessary.

3. Protecting Users' Legal Rights From Anti-Circumvention Law.

    No covered work shall be deemed part of an effective technological
    measure under any applicable law fulfilling obligations under article
    11 of the WIPO copyright treaty adopted on 20 December 1996, or
    similar laws prohibiting or restricting circumvention of such
    measures.
    
    When you convey a covered work, you waive any legal power to forbid
    circumvention of technological measures to the extent such circumvention
    is effected by exercising rights under this License with respect to
    the covered work, and you disclaim any intention to limit operation or
    modification of the work as a means of enforcing, against the work's
    users, your or third parties' legal rights to forbid circumvention of
    technological measures.

4. Conveying Verbatim Copies.

    You may convey verbatim copies of the Program's source code as you
    receive it, in any medium, provided that you conspicuously and
    appropriately publish on each copy an appropriate copyright notice;
    keep intact all notices stating that this License and any
    non-permissive terms added in accord with section 7 apply to the code;
    keep intact all notices of the absence of any warranty; and give all
    recipients a copy of this License along with the Program.
    
    You may charge any price or no price for each copy that you convey,
    and you may offer support or warranty protection for a fee.

5. Conveying Modified Source Versions.

    You may convey a work based on the Program, or the modifications to
    produce it from the Program, in the form of source code under the
    terms of section 4, provided that you also meet all of these conditions:
    
        a) The work must carry prominent notices stating that you modified
        it, and giving a relevant date.
    
        b) The work must carry prominent notices stating that it is
        released under this License and any conditions added under section
        7.  This requirement modifies the requirement in section 4 to
        "keep intact all notices".
    
        c) You must license the entire work, as a whole, under this
        License to anyone who comes into possession of a copy.  This
        License will therefore apply, along with any applicable section 7
        additional terms, to the whole of the work, and all its parts,
        regardless of how they are packaged.  This License gives no
        permission to license the work in any other way, but it does not
        invalidate such permission if you have separately received it.
    
        d) If the work has interactive user interfaces, each must display
        Appropriate Legal Notices; however, if the Program has interactive
        interfaces that do not display Appropriate Legal Notices, your
        work need not make them do so.
    
    A compilation of a covered work with other separate and independent
    works, which are not by their nature extensions of the covered work,
    and which are not combined with it such as to form a larger program,
    in or on a volume of a storage or distribution medium, is called an
    "aggregate" if the compilation and its resulting copyright are not
    used to limit the access or legal rights of the compilation's users
    beyond what the individual works permit.  Inclusion of a covered work
    in an aggregate does not cause this License to apply to the other
    parts of the aggregate.

6. Conveying Non-Source Forms.

    You may convey a covered work in object code form under the terms
    of sections 4 and 5, provided that you also convey the
    machine-readable Corresponding Source under the terms of this License,
    in one of these ways:
    
        a) Convey the object code in, or embodied in, a physical product
        (including a physical distribution medium), accompanied by the
        Corresponding Source fixed on a durable physical medium
        customarily used for software interchange.
    
        b) Convey the object code in, or embodied in, a physical product
        (including a physical distribution medium), accompanied by a
        written offer, valid for at least three years and valid for as
        long as you offer spare parts or customer support for that product
        model, to give anyone who possesses the object code either (1) a
        copy of the Corresponding Source for all the software in the
        product that is covered by this License, on a durable physical
        medium customarily used for software interchange, for a price no
        more than your reasonable cost of physically performing this
        conveying of source, or (2) access to copy the
        Corresponding Source from a network server at no charge.
    
        c) Convey individual copies of the object code with a copy of the
        written offer to provide the Corresponding Source.  This
        alternative is allowed only occasionally and noncommercially, and
        only if you received the object code with such an offer, in accord
        with subsection 6b.
    
        d) Convey the object code by offering access from a designated
        place (gratis or for a charge), and offer equivalent access to the
        Corresponding Source in the same way through the same place at no
        further charge.  You need not require recipients to copy the
        Corresponding Source along with the object code.  If the place to
        copy the object code is a network server, the Corresponding Source
        may be on a different server (operated by you or a third party)
        that supports equivalent copying facilities, provided you maintain
        clear directions next to the object code saying where to find the
        Corresponding Source.  Regardless of what server hosts the
        Corresponding Source, you remain obligated to ensure that it is
        available for as long as needed to satisfy these requirements.
    
        e) Convey the object code using peer-to-peer transmission, provided
        you inform other peers where the object code and Corresponding
        Source of the work are being offered to the general public at no
        charge under subsection 6d.
    
    A separable portion of the object code, whose source code is excluded
    from the Corresponding Source as a System Library, need not be
    included in conveying the object code work.
    
    A "User Product" is either (1) a "consumer product", which means any
    tangible personal property which is normally used for personal, family,
    or household purposes, or (2) anything designed or sold for incorporation
    into a dwelling.  In determining whether a product is a consumer product,
    doubtful cases shall be resolved in favor of coverage.  For a particular
    product received by a particular user, "normally used" refers to a
    typical or common use of that class of product, regardless of the status
    of the particular user or of the way in which the particular user
    actually uses, or expects or is expected to use, the product.  A product
    is a consumer product regardless of whether the product has substantial
    commercial, industrial or non-consumer uses, unless such uses represent
    the only significant mode of use of the product.
    
    "Installation Information" for a User Product means any methods,
    procedures, authorization keys, or other information required to install
    and execute modified versions of a covered work in that User Product from
    a modified version of its Corresponding Source.  The information must
    suffice to ensure that the continued functioning of the modified object
    code is in no case prevented or interfered with solely because
    modification has been made.
    
    If you convey an object code work under this section in, or with, or
    specifically for use in, a User Product, and the conveying occurs as
    part of a transaction in which the right of possession and use of the
    User Product is transferred to the recipient in perpetuity or for a
    fixed term (regardless of how the transaction is characterized), the
    Corresponding Source conveyed under this section must be accompanied
    by the Installation Information.  But this requirement does not apply
    if neither you nor any third party retains the ability to install
    modified object code on the User Product (for example, the work has
    been installed in ROM).
    
    The requirement to provide Installation Information does not include a
    requirement to continue to provide support service, warranty, or updates
    for a work that has been modified or installed by the recipient, or for
    the User Product in which it has been modified or installed.  Access to a
    network may be denied when the modification itself materially and
    adversely affects the operation of the network or violates the rules and
    protocols for communication across the network.
    
    Corresponding Source conveyed, and Installation Information provided,
    in accord with this section must be in a format that is publicly
    documented (and with an implementation available to the public in
    source code form), and must require no special password or key for
    unpacking, reading or copying.

7. Additional Terms.

    "Additional permissions" are terms that supplement the terms of this
    License by making exceptions from one or more of its conditions.
    Additional permissions that are applicable to the entire Program shall
    be treated as though they were included in this License, to the extent
    that they are valid under applicable law.  If additional permissions
    apply only to part of the Program, that part may be used separately
    under those permissions, but the entire Program remains governed by
    this License without regard to the additional permissions.
    
    When you convey a copy of a covered work, you may at your option
    remove any additional permissions from that copy, or from any part of
    it.  (Additional permissions may be written to require their own
    removal in certain cases when you modify the work.)  You may place
    additional permissions on material, added by you to a covered work,
    for which you have or can give appropriate copyright permission.
    
    Notwithstanding any other provision of this License, for material you
    add to a covered work, you may (if authorized by the copyright holders of
    that material) supplement the terms of this License with terms:
    
        a) Disclaiming warranty or limiting liability differently from the
        terms of sections 15 and 16 of this License; or
    
        b) Requiring preservation of specified reasonable legal notices or
        author attributions in that material or in the Appropriate Legal
        Notices displayed by works containing it; or
    
        c) Prohibiting misrepresentation of the origin of that material, or
        requiring that modified versions of such material be marked in
        reasonable ways as different from the original version; or
    
        d) Limiting the use for publicity purposes of names of licensors or
        authors of the material; or
    
        e) Declining to grant rights under trademark law for use of some
        trade names, trademarks, or service marks; or
    
        f) Requiring indemnification of licensors and authors of that
        material by anyone who conveys the material (or modified versions of
        it) with contractual assumptions of liability to the recipient, for
        any liability that these contractual assumptions directly impose on
        those licensors and authors.
    
    All other non-permissive additional terms are considered "further
    restrictions" within the meaning of section 10.  If the Program as you
    received it, or any part of it, contains a notice stating that it is
    governed by this License along with a term that is a further
    restriction, you may remove that term.  If a license document contains
    a further restriction but permits relicensing or conveying under this
    License, you may add to a covered work material governed by the terms
    of that license document, provided that the further restriction does
    not survive such relicensing or conveying.
    
    If you add terms to a covered work in accord with this section, you
    must place, in the relevant source files, a statement of the
    additional terms that apply to those files, or a notice indicating
    where to find the applicable terms.
    
    Additional terms, permissive or non-permissive, may be stated in the
    form of a separately written license, or stated as exceptions;
    the above requirements apply either way.

8. Termination.

    You may not propagate or modify a covered work except as expressly
    provided under this License.  Any attempt otherwise to propagate or
    modify it is void, and will automatically terminate your rights under
    this License (including any patent licenses granted under the third
    paragraph of section 11).
    
    However, if you cease all violation of this License, then your
    license from a particular copyright holder is reinstated (a)
    provisionally, unless and until the copyright holder explicitly and
    finally terminates your license, and (b) permanently, if the copyright
    holder fails to notify you of the violation by some reasonable means
    prior to 60 days after the cessation.
    
    Moreover, your license from a particular copyright holder is
    reinstated permanently if the copyright holder notifies you of the
    violation by some reasonable means, this is the first time you have
    received notice of violation of this License (for any work) from that
    copyright holder, and you cure the violation prior to 30 days after
    your receipt of the notice.
    
    Termination of your rights under this section does not terminate the
    licenses of parties who have received copies or rights from you under
    this License.  If your rights have been terminated and not permanently
    reinstated, you do not qualify to receive new licenses for the same
    material under section 10.

9. Acceptance Not Required for Having Copies.

    You are not required to accept this License in order to receive or
    run a copy of the Program.  Ancillary propagation of a covered work
    occurring solely as a consequence of using peer-to-peer transmission
    to receive a copy likewise does not require acceptance.  However,
    nothing other than this License grants you permission to propagate or
    modify any covered work.  These actions infringe copyright if you do
    not accept this License.  Therefore, by modifying or propagating a
    covered work, you indicate your acceptance of this License to do so.

10. Automatic Licensing of Downstream Recipients.

    Each time you convey a covered work, the recipient automatically
    receives a license from the original licensors, to run, modify and
    propagate that work, subject to this License.  You are not responsible
    for enforcing compliance by third parties with this License.
    
    An "entity transaction" is a transaction transferring control of an
    organization, or substantially all assets of one, or subdividing an
    organization, or merging organizations.  If propagation of a covered
    work results from an entity transaction, each party to that
    transaction who receives a copy of the work also receives whatever
    licenses to the work the party's predecessor in interest had or could
    give under the previous paragraph, plus a right to possession of the
    Corresponding Source of the work from the predecessor in interest, if
    the predecessor has it or can get it with reasonable efforts.
    
    You may not impose any further restrictions on the exercise of the
    rights granted or affirmed under this License.  For example, you may
    not impose a license fee, royalty, or other charge for exercise of
    rights granted under this License, and you may not initiate litigation
    (including a cross-claim or counterclaim in a lawsuit) alleging that
    any patent claim is infringed by making, using, selling, offering for
    sale, or importing the Program or any portion of it.

11. Patents.

    A "contributor" is a copyright holder who authorizes use under this
    License of the Program or a work on which the Program is based.  The
    work thus licensed is called the contributor's "contributor version".
    
    A contributor's "essential patent claims" are all patent claims
    owned or controlled by the contributor, whether already acquired or
    hereafter acquired, that would be infringed by some manner, permitted
    by this License, of making, using, or selling its contributor version,
    but do not include claims that would be infringed only as a
    consequence of further modification of the contributor version.  For
    purposes of this definition, "control" includes the right to grant
    patent sublicenses in a manner consistent with the requirements of
    this License.
    
    Each contributor grants you a non-exclusive, worldwide, royalty-free
    patent license under the contributor's essential patent claims, to
    make, use, sell, offer for sale, import and otherwise run, modify and
    propagate the contents of its contributor version.
    
    In the following three paragraphs, a "patent license" is any express
    agreement or commitment, however denominated, not to enforce a patent
    (such as an express permission to practice a patent or covenant not to
    sue for patent infringement).  To "grant" such a patent license to a
    party means to make such an agreement or commitment not to enforce a
    patent against the party.
    
    If you convey a covered work, knowingly relying on a patent license,
    and the Corresponding Source of the work is not available for anyone
    to copy, free of charge and under the terms of this License, through a
    publicly available network server or other readily accessible means,
    then you must either (1) cause the Corresponding Source to be so
    available, or (2) arrange to deprive yourself of the benefit of the
    patent license for this particular work, or (3) arrange, in a manner
    consistent with the requirements of this License, to extend the patent
    license to downstream recipients.  "Knowingly relying" means you have
    actual knowledge that, but for the patent license, your conveying the
    covered work in a country, or your recipient's use of the covered work
    in a country, would infringe one or more identifiable patents in that
    country that you have reason to believe are valid.
    
    If, pursuant to or in connection with a single transaction or
    arrangement, you convey, or propagate by procuring conveyance of, a
    covered work, and grant a patent license to some of the parties
    receiving the covered work authorizing them to use, propagate, modify
    or convey a specific copy of the covered work, then the patent license
    you grant is automatically extended to all recipients of the covered
    work and works based on it.
    
    A patent license is "discriminatory" if it does not include within
    the scope of its coverage, prohibits the exercise of, or is
    conditioned on the non-exercise of one or more of the rights that are
    specifically granted under this License.  You may not convey a covered
    work if you are a party to an arrangement with a third party that is
    in the business of distributing software, under which you make payment
    to the third party based on the extent of your activity of conveying
    the work, and under which the third party grants, to any of the
    parties who would receive the covered work from you, a discriminatory
    patent license (a) in connection with copies of the covered work
    conveyed by you (or copies made from those copies), or (b) primarily
    for and in connection with specific products or compilations that
    contain the covered work, unless you entered into that arrangement,
    or that patent license was granted, prior to 28 March 2007.
    
    Nothing in this License shall be construed as excluding or limiting
    any implied license or other defenses to infringement that may
    otherwise be available to you under applicable patent law.

12. No Surrender of Others' Freedom.

    If conditions are imposed on you (whether by court order, agreement or
    otherwise) that contradict the conditions of this License, they do not
    excuse you from the conditions of this License.  If you cannot convey a
    covered work so as to satisfy simultaneously your obligations under this
    License and any other pertinent obligations, then as a consequence you may
    not convey it at all.  For example, if you agree to terms that obligate you
    to collect a royalty for further conveying from those to whom you convey
    the Program, the only way you could satisfy both those terms and this
    License would be to refrain entirely from conveying the Program.

13. Use with the GNU Affero General Public License.

    Notwithstanding any other provision of this License, you have
    permission to link or combine any covered work with a work licensed
    under version 3 of the GNU Affero General Public License into a single
    combined work, and to convey the resulting work.  The terms of this
    License will continue to apply to the part which is the covered work,
    but the special requirements of the GNU Affero General Public License,
    section 13, concerning interaction through a network will apply to the
    combination as such.

14. Revised Versions of this License.

    The Free Software Foundation may publish revised and/or new versions of
    the GNU General Public License from time to time.  Such new versions will
    be similar in spirit to the present version, but may differ in detail to
    address new problems or concerns.
    
    Each version is given a distinguishing version number.  If the
    Program specifies that a certain numbered version of the GNU General
    Public License "or any later version" applies to it, you have the
    option of following the terms and conditions either of that numbered
    version or of any later version published by the Free Software
    Foundation.  If the Program does not specify a version number of the
    GNU General Public License, you may choose any version ever published
    by the Free Software Foundation.
    
    If the Program specifies that a proxy can decide which future
    versions of the GNU General Public License can be used, that proxy's
    public statement of acceptance of a version permanently authorizes you
    to choose that version for the Program.
    
    Later license versions may give you additional or different
    permissions.  However, no additional obligations are imposed on any
    author or copyright holder as a result of your choosing to follow a
    later version.

15. Disclaimer of Warranty.

    [bold]THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
    APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
    HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
    OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
    IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
    ALL NECESSARY SERVICING, REPAIR OR CORRECTION.[/]

16. Limitation of Liability.

    [bold]IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
    WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
    THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
    GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
    USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
    DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
    PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
    EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGES.[/]

17. Interpretation of Sections 15 and 16.

    If the disclaimer of warranty and limitation of liability provided
    above cannot be given local legal effect according to their terms,
    reviewing courts shall apply local law that most closely approximates
    an absolute waiver of all civil liability in connection with the
    Program, unless a warranty or assumption of liability accompanies a
    copy of the Program in return for a fee.

[bold]END OF TERMS AND CONDITIONS[/]
"""

    notice: str = f"""
{_project.name} {Version.release} Copyright © {_current_year} {rich_colours.BOLD_BLUE}[link={_author.gravatar}]{_author.name}[/link]{rich_colours.BOLD_BLUE_RESET}

This program comes with [bold]ABSOLUTELY NO WARRANTY[/]; for details, use {rich_colours.BLUE}licence -w/--warranty{rich_colours.BLUE_RESET}.
This is free software, and you are welcome to redistribute it
[bold]UNDER CERTAIN CONDITIONS[/]; use {rich_colours.BLUE}license -c/--conditions{rich_colours.BLUE_RESET} for details.
"""

    warranty: str = f"""
Reddit-data analysis toolkit.
Copyright © {_current_year} {rich_colours.BOLD_BLUE}[link={_author.gravatar}]{_author.name}[/link]{rich_colours.BOLD_BLUE_RESET}

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but [bold]WITHOUT ANY WARRANTY[/]; without even the implied warranty of
[bold]MERCHANTABILITY[/] or [bold]FITNESS FOR A PARTICULAR PURPOSE[/].  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


# -------------------------------- END ----------------------------------------- #



import inspect
import os
import typing as t
from datetime import datetime

import requests
import rich_click as click
from karmakrate.everything.runtime_things import RuntimeThings
from karmakrate.handlers.io_handlers import DataFrameHandler, FileHandler
from karmakrate.riches import rich_colours
from karmakrate.riches.rich_logging import console, logger
from karmakrate.riches.rich_render import Render
from prawcore import exceptions
from rich.status import Status

from ..meta.about import Project
from ..meta.version import Version

__all__ = ["run"]

NORMAL_ERROR_PREFIX: str = f"{rich_colours.RED}⚠{rich_colours.RED_RESET}"
CRITICAL_ERROR_PREFIX: str = f"{rich_colours.BOLD_RED}⚠{rich_colours.BOLD_RED_RESET}"
WARNING_PREFIX: str = f"{rich_colours.BOLD_YELLOW}⚠{rich_colours.BOLD_YELLOW_RESET}"


def invoke_method(
    method: t.Callable,
    **kwargs: t.Union[str, click.Context, requests.Session, Status],
):
    """
    Invoke a method with the keyword arguments it accepts, and optionally export the result.

    :param method: The method to invoke.
    :type method: Callable
    :param kwargs: Keyword arguments that may include:
        - ``export`` (str): Comma-separated list of export formats (e.g., "csv,json").
        - ``status`` (Status): A Rich status indicator.
        - ``session`` (requests.Session): An HTTP session to use.
        - ``argument`` (str): The argument that triggered the method.
        - ``ctx`` (click.Context): The Click context object.
        - ``logger`` (logging.Logger): Optional logger.

    :raises ValueError: If the method cannot be called due to argument mismatch or invalid data.

    :return: None
    """
    ctx: click.Context = kwargs.get("ctx")
    session = kwargs.get("session")
    status = kwargs.get("status")

    command: str = ctx.command.name
    argument: str = kwargs.get("argument")

    status.update(f"Initialising {ctx.command.name} module...")
    # 🔍 Filter out only those kwargs that the method actually accepts
    sig = inspect.signature(method)
    accepted_kwargs = {
        key: value for key, value in kwargs.items() if key in sig.parameters
    }

    # 👇 Add direct assignment variables to the accepted args if needed
    if "session" in sig.parameters:
        accepted_kwargs["session"] = session
    if "status" in sig.parameters:
        accepted_kwargs["status"] = status
    if "logger" in sig.parameters:
        accepted_kwargs["logger"] = kwargs.get("logger")

    # 🧠 Actually call the method
    response_data: t.Union[t.List, t.Dict, str, bool, t.Any] = method(**accepted_kwargs)

    if response_data:
        Render.panels(data=response_data)
        if kwargs.get("export"):
            exports_child_dir: str = os.path.join(
                FileHandler.PARENT_DIR,
                "exports",
                command,
                argument,
            )

            FileHandler.pathfinder(
                directories=[
                    os.path.join(exports_child_dir, extension)
                    for extension in ["csv", "html", "json", "xml"]
                ],
            )

            export_to: t.List[str] = kwargs.get("export").split(",")
            dataframe = DataFrameHandler.build(data=response_data, status=status)
            DataFrameHandler.export(
                dataframe=dataframe,
                filename=FileHandler.time_to_filename(),
                directory=exports_child_dir,
                formats=export_to,
                status=status,
            )


def route_to_method(
    ctx: click.Context,
    method_map: t.Dict,
    export: str,
    **kwargs: t.Union[str, int, bool],
):
    """
    Routes CLI arguments to the appropriate method and delegates execution to `invoke_method`.

    :param ctx: The current Click context.
    :type ctx: click.Context
    :param method_map: A mapping of CLI argument names to their corresponding functions.
    :type method_map: Dict[str, Callable]
    :param export: Comma-separated string indicating desired export formats (e.g. "csv,html").
    :type export: str
    :param kwargs: Parsed CLI arguments (from `click`) that may contain trigger flags for each method.
    :type kwargs: Union[str, int, bool]

    Side Effects:
        - Prints status updates and errors to the riches.
        - Displays a license notice.
        - Performs data export if applicable.
        - Logs execution details and exceptions.

    If no valid argument is provided, prints command usage help.
    """
    runtime_operations = RuntimeThings(
        package_name=Project.package, version_cls=Version
    )

    is_valid_arg: bool = False

    for argument, method in method_map.items():
        if kwargs.get(argument):
            is_valid_arg = True
            start_time: datetime = datetime.now()
            try:
                runtime_operations.clear_screen()
                with Status(
                    status=f"Starting",
                    console=console,
                ) as status:
                    with requests.Session() as session:
                        runtime_operations.check_updates(session=session, status=status)
                        runtime_operations.check_status(session=session, status=status)

                    invoke_method(
                        method=method,
                        status=status,
                        ctx=ctx,
                        export=export,
                        argument=argument,
                    )
            except exceptions.TooManyRequests as too_many_requests:
                logger.warning(
                    f"{WARNING_PREFIX} Woah! Chill out, dude: {too_many_requests}"
                )
            except exceptions.BadRequest as bad_request:
                logger.error(
                    f"{NORMAL_ERROR_PREFIX} A BadRequest error occurred: {bad_request}"
                )
            except exceptions.ServerError as server_error:
                logger.error(
                    f"{NORMAL_ERROR_PREFIX} A ServerError occurred: {server_error}"
                )
            except exceptions.RequestException as response_exception:
                logger.error(
                    f"{NORMAL_ERROR_PREFIX} A ResponseException error occurred: {response_exception}"
                )
            except exceptions.PrawcoreException as prawcore_exception:
                logger.critical(
                    f"{CRITICAL_ERROR_PREFIX} A PrawcoreException error: {prawcore_exception}"
                )
            except Exception as error:
                logger.critical(
                    f"{CRITICAL_ERROR_PREFIX} An unexpected error occurred: {error}"
                )
            finally:
                console.print(
                    f":keyboard: {rich_colours.BOLD_BLUE}[link=https://github.com/{Project.package}-io]GitHub[/link]{rich_colours.BOLD_BLUE_RESET}"
                    " | "
                    f":books: {rich_colours.BOLD_BLUE}[link=https://{Project.package}.readthedocs.io]Documentation[/link]{rich_colours.BOLD_BLUE_RESET}"
                    " | "
                    f":black_heart: {rich_colours.BOLD_BLUE}[link=https://opencollective.com/{Project.package}]Become a Sponsor[/link]{rich_colours.BOLD_BLUE_RESET}",
                    justify="center",
                    style=rich_colours.BOLD_WHITE.strip("[,]"),
                )

    if not is_valid_arg:
        ctx.command.get_usage(ctx=ctx)


def run(
    ctx: click.Context,
    export: str,
    method_map: t.Dict[str, t.Callable],
    **kwargs: t.Any,
):
    """
    Entrypoint to execute a modular CLI command.

    :param ctx: The Click context for the current CLI invocation.
    :type ctx: click.Context
    :param export: Comma-separated string of export formats (e.g., "csv,json").
    :type export: str
    :param method_map: Dictionary mapping CLI argument names to callable functions.
    :type method_map: Dict[str, Callable]
    :param kwargs: Additional keyword arguments passed from Click options/arguments.
    :type kwargs: Any
    """
    route_to_method(
        ctx=ctx,
        method_map=method_map,
        export=export,
        **kwargs,
    )



import functools
import typing as t

import rich_click as click
from karmakrate.everything.runtime_things import RuntimeThings
from karmakrate.handlers.auth_handler import AuthHandler
from karmakrate.riches.rich_logging import console

from .main import run
from ..core.client import LISTINGS, SORT, TIME_FILTERS
from ..core.post import Post
from ..core.posts import Posts
from ..core.search import Search
from ..core.subreddit import Subreddit
from ..core.subreddits import Subreddits
from ..core.user import User
from ..core.users import Users
from ..meta.about import Project
from ..meta.license import License
from ..meta.version import Version

__all__ = ["start"]


def set_window_title(text: t.Optional[str] = None):
    title = f"{Project.name} v{Version.release}"

    if text:
        title = f"{title} - {text}"

    console.set_window_title(title)


def global_options(func: t.Callable) -> t.Callable:
    """
    Decorator to add global CLI options like sort, timeframe, export, etc.,
    and inject them into ctx.obj for use within the command.
    """

    @click.option(
        "-e",
        "--export",
        type=str,
        help="A comma-separated list <w/o whitespaces> of file types to export the output to <supported: csv,html,json,xml>",
    )
    @click.option(
        "--listing",
        default="top",
        show_default=True,
        type=click.Choice(t.get_args(LISTINGS)),
        help="Reddit Listings (I'm still working on a better description)",
    )
    @click.option(
        "-l",
        "--limit",
        default=100,
        show_default=True,
        type=int,
        help="Maximum data output limit <max 100 if searching for users>",
    )
    @click.option(
        "-s",
        "--sort",
        default="lucene",
        show_default=True,
        type=click.Choice(t.get_args(SORT)),
        help="Sort criterion",
    )
    @click.option(
        "-t",
        "--time-filter",
        default="all",
        show_default=True,
        type=click.Choice(t.get_args(TIME_FILTERS)),
        help="Filter data by time",
    )
    @click.pass_context
    @functools.wraps(func)
    def wrapper(
        ctx: click.Context,
        time_filter: TIME_FILTERS,
        sort: SORT,
        limit: int,
        export: str,
        listing: str,
        *args,
        **kwargs,
    ):
        ctx.ensure_object(dict)
        ctx.obj["time_filter"] = time_filter
        ctx.obj["sort"] = sort
        ctx.obj["limit"] = limit
        ctx.obj["export"] = export
        ctx.obj["listing"] = listing
        return ctx.invoke(func, *args, **kwargs)

    return wrapper


def help_callback(ctx: click.Context, _, value: bool):
    """
    Custom callback function for handling the '--help' option in Click commands.
    """

    if value and not ctx.resilient_parsing:
        click.echo(ctx.get_help())
        if RuntimeThings.is_snap_package():
            click.pause()
        ctx.exit()


@click.group(
    help=f"""
{Project.summary}\n\n\n{Project.description}""",
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option(
    "-h",
    "--help",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=help_callback,
    help="Show this message and exit.",
)
@click.version_option(
    Version.full_version,
    "-v",
    "--version",
)
@global_options
@click.pass_context
def cli(
    ctx: click.Context,
):
    """
    Main CLI group for Knew Karma.
    """
    set_window_title()
    ctx.ensure_object(dict)


@cli.command(name="auth", help="Authenticate with Reddit")
@click.option("--client-id", help="Reddit API client id", type=str)
@click.option("--client-secret", help="Reddit API client secret", type=str)
def auth(client_id: t.Optional[str], client_secret: t.Optional[str]):
    AuthHandler.write(client_id=client_id, client_secret=client_secret)


@cli.command("license")
@click.option("-c", "--conditions", help="License terms and conditions.", is_flag=True)
@click.option("-w", "--warranty", help="License warranty.", is_flag=True)
@click.pass_context
def licence(
    ctx: click.Context, conditions: t.Optional[bool], warranty: t.Optional[bool]
):
    """
    Show license information
    """
    set_window_title(
        "License Terms and Conditions"
        if conditions
        else "License Warranty" if warranty else None
    )
    if conditions:
        console.print(
            License.conditions,
            justify="center",
        )
    elif warranty:
        console.print(
            License.warranty,
            justify="center",
        )
    else:
        click.echo(ctx.get_help())


@cli.command(
    name="post",
    help="Use this command to get an individual post's data including its comments.",
)
@click.argument("id")
@click.option("--info", is_flag=True, help="Get post info (w/o comments)")
@click.option("--comments", is_flag=True, help="Get post comments")
@global_options
@click.pass_context
def post(ctx: click.Context, id: str, info: bool, comments: bool):
    export: str = ctx.obj["export"]

    r_post = Post(id=id)
    method_map: t.Dict = {
        "comments": lambda session, status, logger: r_post.comments(status=status),
        "info": lambda session, status, logger: r_post.info(status=status),
    }

    run(
        ctx=ctx,
        method_map=method_map,
        export=export,
        info=info,
        comments=comments,
    )


@cli.command(
    name="posts",
    help="Use this command get best, controversial, front-page, new, popular, and/or rising posts.",
)
@click.option("--best", is_flag=True, help="Get posts from the best listing")
@click.option(
    "--controversial",
    is_flag=True,
    help="Get posts from the controversial listing",
)
@click.option(
    "--front-page",
    is_flag=True,
    help="Get posts from the reddit front-page",
)
@click.option("--new", is_flag=True, help="Get new posts")
@click.option(
    "--top",
    is_flag=True,
    help="Get posts from the top listing",
)
@click.option("--rising", is_flag=True, help="Get posts from the rising listing")
@global_options
@click.pass_context
def cmd_posts(
    ctx: click.Context,
    best: bool,
    controversial: bool,
    front_page: bool,
    new: bool,
    top: bool,
    rising: bool,
):
    """
    Retrieve various types of posts such as best, controversial, popular, new, and front-page.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param best: Flag to get posts from the best listing.
    :type best: bool
    :param controversial: Flag to get posts from the controversial listing.
    :type controversial: bool
    :param front_page: Flag to get posts from the reddit front-page.
    :type front_page: bool
    :param new: Flag to get new posts.
    :type new: bool
    param top: Flag to get posts from the top listing.
    :type top: bool
    :param rising: Flag to get posts from the rising listing.
    :type rising: bool
    """

    time_filter: TIME_FILTERS = ctx.obj["time_filter"]
    sort: SORT = ctx.obj["sort"]
    limit: int = ctx.obj["limit"]
    export: str = ctx.obj["export"]

    method_map: t.Dict = {
        "best": lambda session, status, logger: Posts.best(
            limit=limit,
            status=status,
            session=session,
        ),
        "controversial": lambda session, status, logger: Posts.controversial(
            limit=limit,
            status=status,
            session=session,
        ),
        "front_page": lambda session, status, logger: Posts.front_page(
            limit=limit, status=status, session=session
        ),
        "new": lambda session, status, logger: Posts.new(
            limit=limit, status=status, session=session
        ),
        "top": lambda session, status, logger: Posts.top(
            limit=limit,
            status=status,
            session=session,
        ),
        "rising": lambda session, status, logger: Posts.rising(
            limit=limit, status=status, session=session
        ),
    }

    run(
        ctx=ctx,
        method_map=method_map,
        export=export,
        best=best,
        controversial=controversial,
        front_page=front_page,
        new=new,
        top=top,
        rising=rising,
    )


@cli.command(
    name="search",
    help="Use this command for search/discovery of users, subreddits, and posts.",
)
@click.argument("query")
@click.option("--posts", is_flag=True, help="Search posts")
@click.option("--subreddits", is_flag=True, help="Search subreddits")
@click.option("--users", is_flag=True, help="Search users")
@global_options
@click.pass_context
def cmd_search(
    ctx: click.Context, query: str, posts: bool, subreddits: bool, users: bool
):
    """
    Search for posts, subreddits, or users based on a query.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param query: The search query.
    :type query: str
    :param posts: Flag to search posts.
    :type posts: bool
    :param subreddits: Flag to search subreddits.
    :type subreddits: bool
    :param users: Flag to search users.
    :type users: bool
    """

    sort: SORT = ctx.obj["sort"]
    limit: int = ctx.obj["limit"]
    export: str = ctx.obj["export"]

    search = Search(
        query=query,
    )
    method_map: t.Dict = {
        "posts": lambda session, status, logger: search.posts(
            limit=limit, status=status, session=session
        ),
        "subreddits": lambda session, status, logger: search.subreddits(
            limit=limit, session=session, status=status
        ),
        "users": lambda session, status, logger: search.users(
            limit=limit, status=status, session=session
        ),
    }

    run(
        ctx=ctx,
        method_map=method_map,
        export=export,
        posts=posts,
        subreddits=subreddits,
        users=users,
    )


@cli.command(
    name="user",
    help="Use this command to get user data, such as profile, posts, "
    "comments, top subreddits, moderated subreddits, and more...",
)
@click.argument("username")
@click.option("--comments", is_flag=True, help="Get comments")
@click.option(
    "--moderated",
    is_flag=True,
    help="Get moderated subreddits",
)
@click.option("--overview", is_flag=True, help="Get most recent comments")
@click.option("--posts", is_flag=True, help="Get posts")
@click.option("--profile", is_flag=True, help="Get profile")
@click.option(
    "--top-subreddits",
    type=int,
    help="Get top n subreddits",
)
@global_options
@click.pass_context
def user(
    ctx: click.Context,
    username: str,
    comments: bool,
    moderated: bool,
    overview: bool,
    posts: bool,
    profile: bool,
    top_subreddits: int,
):
    time_filter: TIME_FILTERS = ctx.obj["time_filter"]
    limit: int = ctx.obj["limit"]
    export: str = ctx.obj["export"]
    listing: LISTINGS = ctx.obj["listing"]

    r_user = User(username=username)
    method_map: t.Dict = {
        "comments": lambda status, logger: r_user.comments(
            limit=limit, listing=listing, status=status
        ),
        "moderated": lambda status, logger: r_user.moderated(status=status),
        "overview": lambda status, logger: r_user.overview(status=status),
        "posts": lambda status, logger: r_user.posts(
            limit=limit, listing=listing, status=status
        ),
        "profile": lambda status, logger: r_user.profile(status=status),
        "top_subreddits": lambda status, logger: r_user.top_subreddits(
            top_n=top_subreddits, status=status
        ),
    }

    run(
        ctx=ctx,
        method_map=method_map,
        export=export,
        comments=comments,
        moderated=moderated,
        overview=overview,
        posts=posts,
        profile=profile,
        top_subreddits=top_subreddits,
    )


@cli.command(
    name="users",
    help="Use this command to get all, new, and/or popular users.",
)
@click.option("-a", "--all", "_all", is_flag=True, help="Get all users")
@click.option(
    "-n",
    "--new",
    is_flag=True,
    help="Get new users",
)
@click.option(
    "-p",
    "--popular",
    is_flag=True,
    help="Get popular users",
)
@global_options
@click.pass_context
def cmd_users(ctx: click.Context, _all: bool, new: bool, popular: bool):
    """
    Retrieve various users such as new, popular, and all users.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param _all: Flag to get all users.
    :type _all: bool
    :param new: Flag to get new users.
    :type new: bool
    :param popular: Flag to get popular users.
    :type popular: bool
    """

    export: str = ctx.obj["export"]
    time_filter: TIME_FILTERS = ctx.obj["time_filter"]
    limit: int = ctx.obj["limit"]

    method_map: t.Dict = {
        "all": lambda session, status, logger: Users.all(
            session=session,
            limit=limit,
            status=status,
        ),
        "new": lambda session, status, logger: Users.new(
            session=session,
            limit=limit,
            status=status,
        ),
        "popular": lambda session, status, logger: Users.popular(
            session=session,
            limit=limit,
            status=status,
        ),
    }

    run(
        ctx=ctx,
        method_map=method_map,
        export=export,
        all=all,
        new=new,
        popular=popular,
    )


@cli.command(
    name="subreddit",
    help="Get data from a specified subreddit.",
)
@click.argument("display_name")
@click.option("--comments", is_flag=True, type=str, help="Get comments")
@click.option("--posts", is_flag=True, help="Get posts")
@click.option("--profile", is_flag=True, help="Get profile")
@click.option("--search", type=str, help="Search for posts that match a query")
@click.option("--wiki-pages", is_flag=True, help="Get wiki pages")
@global_options
@click.pass_context
def subreddit(
    ctx: click.Context,
    display_name: str,
    comments: bool,
    posts: bool,
    profile: bool,
    search: str,
    wiki_pages: bool,
):
    time_filter: TIME_FILTERS = ctx.obj["time_filter"]
    sort: SORT = ctx.obj["sort"]
    limit: int = ctx.obj["limit"]
    export: str = ctx.obj["export"]
    listing: LISTINGS = ctx.obj["listing"]

    r_subreddit = Subreddit(display_name=display_name)
    method_map = {
        "comments": lambda status, logger: r_subreddit.comments(
            limit=limit, status=status
        ),
        "posts": lambda status, logger: r_subreddit.posts(
            limit=limit, listing=listing, status=status
        ),
        "profile": lambda status, logger: r_subreddit.profile(status=status),
        "search": lambda status, logger: r_subreddit.search(
            query=search,
            limit=limit,
            sort=sort,
            time_filter=time_filter,
            status=status,
        ),
        "wiki_pages": lambda status, logger: r_subreddit.wiki_pages(status=status),
    }
    run(
        ctx=ctx,
        method_map=method_map,
        export=export,
        comments=comments,
        profile=profile,
        posts=posts,
        search=search,
        wiki_pages=wiki_pages,
        listing=listing,
    )


@cli.command(
    name="subreddits",
    help="Use this command to get all, default, new, and/or popular subreddits.",
)
@click.option("-a", "--all", "_all", is_flag=True, help="Get all subreddits")
@click.option(
    "-d",
    "--default",
    is_flag=True,
    help="Get default subreddits",
)
@click.option(
    "-n",
    "--new",
    is_flag=True,
    help="Get new subreddits",
)
@click.option(
    "-p",
    "--popular",
    is_flag=True,
    help="Get popular subreddits",
)
@global_options
@click.pass_context
def cmd_subreddits(
    ctx: click.Context, _all: bool, default: bool, new: bool, popular: bool
):
    """
    Retrieve various subreddits such as new, popular, default, and all subreddits.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param _all: Flag to get all subreddits.
    :type _all: bool
    :param default: Flag to get default subreddits.
    :type default: bool
    :param new: Flag to get new subreddits.
    :type new: bool
    :param popular: Flag to get popular subreddits.
    :type popular: bool
    """

    export: str = ctx.obj["export"]
    limit: int = ctx.obj["limit"]

    method_map: t.Dict = {
        "all": lambda session, status, logger: Subreddits.all(
            limit=limit,
            session=session,
            status=status,
        ),
        "default": lambda session, status, logger: Subreddits.default(
            limit=limit, session=session, status=status
        ),
        "new": lambda session, status, logger: Subreddits.new(
            limit=limit, session=session, status=status
        ),
        "popular": lambda session, status, logger: Subreddits.popular(
            limit=limit, session=session, status=status
        ),
    }

    run(
        ctx=ctx,
        method_map=method_map,
        export=export,
        all=all,
        default=default,
        new=new,
        popular=popular,
    )


def start():
    """
    Main entrypoint for the Knew Karma command-line interface.
    """

    cli(obj={})


# -------------------------------- END ----------------------------------------- #



import typing as t

import requests
from praw.models import Redditor
from rich.status import Status

from .client import reddit


class Users:
    """Represents Reddit users and provides methods for getting related data."""

    @classmethod
    def new(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Redditor]:
        new_users = reddit.users(
            session=session,
            status=status,
            kind="new",
            limit=limit,
        )

        return new_users

    @classmethod
    def popular(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Redditor]:
        popular_users = reddit.users(
            session=session,
            status=status,
            kind="popular",
            limit=limit,
        )

        return popular_users

    @classmethod
    def all(
        cls,
        session: requests.Session,
        limit: int,
        status: t.Optional[Status] = None,
    ) -> t.List[Redditor]:
        all_users = reddit.users(
            session=session,
            status=status,
            kind="all",
            limit=limit,
        )

        return all_users
